#include <SoftwareSerial.h>
#include <toneAC.h>

const int BUFFER_SIZE = 14; // RFID DATA FRAME FORMAT: 1byte head (value: 2), 10 byte data (2 byte version + 8 byte tag), 2byte checksum, 1byte tail (value: 3).
SoftwareSerial ssrfid = SoftwareSerial(A2, A1); 

uint8_t buffer[BUFFER_SIZE]; // used to store an incoming data frame 
int buffer_index = 0;
unsigned long lastTagNumber;
unsigned long tagNumber;
unsigned long lastTime;

// Initialize the device, fire up the red LED and play some welcome tune.
void setup(){
  lastTagNumber = 0;
  lastTime = 0;
  Serial.begin(9600); 
  ssrfid.begin(9600);
  ssrfid.listen(); 
  pinMode(2, OUTPUT);
  pinMode(3, OUTPUT);
  digitalWrite(2, HIGH);
  digitalWrite(3, LOW);
  // Play intro
  toneAC(660, 10, 200);
  delay(100);
  toneAC(660, 10, 100);
  delay(50);
  toneAC(660, 10, 100);
  delay(50);
  toneAC(660, 10, 200);
  delay(100);
  toneAC(0);
  // Serial.println("---READY--");
}

// Wait (forever) for any tags to be recognized.
void loop(){
  if(ssrfid.available()>0){
    bool call_extract_tag = false;
    int ssvalue;
    ssvalue = ssrfid.read(); 
    if(ssvalue == -1){ // No data available.
      return;
    }
    if(ssvalue == 2) { // RDM6300 found a tag => tag ID incoming. 
      buffer_index = 0;
    }
    else if(ssvalue == 3) { // Tag has been fully transmitted.
      call_extract_tag = true; // Extract tag at the end of the loop.
    }
    if(buffer_index >= BUFFER_SIZE) { // Checking for a buffer overflow (It's very unlikely that an buffer overflow comes up!)
      // Buffer overflow detected. Send a zero code to inform the host that this happened.
      // (You may delete this if this is really happens and if it is an obstacle for the host).
      Serial.println("00000000");
      return;
    }
    buffer[buffer_index++] = ssvalue; // Everything is alright => copy current value to buffer.
    if(call_extract_tag == true) {
      if(buffer_index == BUFFER_SIZE) {
        long tag = extract_tag();
      }
      else{ // Something is wrong... start again looking for preamble (value: 2).
        buffer_index = 0;
        return;
      }
    }    
  }    
}

// Convert a hex string to a number (= long integer).
long asciiToNumber(uint8_t digit){
  long number;
  if((digit >= 65)&&(digit <= 70)){ // A-F.
    number = digit - 55;
  }
  else if((digit >= 48)&&(digit <= 57)){
    number = digit - 48;
  }
  else{
    return -1;
  }
  return number;
}

// Extract the tag from the information sent by the RDM6300.
long extract_tag(){
  long checkSumCalculated, checkSumFromData;
  long checkSumByte;
  long number;
  long versionNumber;
  unsigned long tagNumberCopy;
  char tagNumberWithLeadingZeros[10];
  int i;
  
  // A) Check start and end byte for format errors.
  if((buffer[0] != 2) // First byte: must be 2.
    ||(buffer[13] != 3)){ // Last byte: must be 3.
    return -1; // Wrong tag format.
  }
  // B) Check the version.
  versionNumber = 0;
  for(i = 0; i < 2; i++){
    number = asciiToNumber(buffer[1 + i]);
    versionNumber = versionNumber<<4;
    versionNumber += number;
  }
  // Use the following lines if you want to ensure that the tags/cards you are using
  // are from specific batches.
//  if((versionNumber != 19)&&(versionNumber != 67)){
//    return -1; // Wrong key version.
//  }
  // C) Extract the tag number.
  tagNumber = 0;
  for(i = 0; i < 8; i++){
    number = asciiToNumber(buffer[3 + i]);
    if(number == -1){
      return -1; // Format error.
    }
    tagNumber = tagNumber<<4;
    tagNumber += number;
  }
  // C) Check checksum.
  checkSumFromData = asciiToNumber(buffer[11]);
  checkSumFromData = checkSumFromData<<4;
  checkSumFromData += asciiToNumber(buffer[12]);
  checkSumCalculated = versionNumber;
  tagNumberCopy = tagNumber;
  for(i = 0; i < 4; i++){
    checkSumByte = tagNumberCopy & 0xFF;
    checkSumCalculated ^= checkSumByte;
    tagNumberCopy = tagNumberCopy >> 8;
  }
  if(checkSumCalculated != checkSumFromData){
    return -1;
  }
  // Ensure that the same tag/card does not send a code to the host within 2 seconds twice.
  if(millis() > lastTime + 2000){
    lastTagNumber = 0; // Reset stored tag number after 2 seconds.
  }
  if(tagNumber != lastTagNumber){
    ssrfid.end(); // Stop the transmission for aviding interefences of the serial communication (= interrupts) with the tone PWM.
    // Send chip ID as a 10 digit decimal integer with leading zeros + '\r\n' which will be appended by Serial.println.
    sprintf(tagNumberWithLeadingZeros, "%010ld", tagNumber);
    Serial.println(tagNumberWithLeadingZeros);
    // Play "fanfare". 
    digitalWrite(2, LOW);
    digitalWrite(3, HIGH);
    toneAC(330, 10, 150);
    toneAC(415, 10, 150);
    toneAC(494, 10, 150);
    toneAC(660, 10, 300);
    delay(200);
    toneAC(0);
    digitalWrite(2, HIGH);
    digitalWrite(3, LOW);
    lastTagNumber = tagNumber;
    lastTime = millis();
    ssrfid.begin(9600); // Re-start the transmission.
    ssrfid.listen(); 
  }
  // D) Optional: Send the full tag number as hex-code in ASCII.
  // buffer[11] = 0;
  // Serial.println((char*)&buffer[3]);
  return 0;
}
